/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/signal.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>

#include <linux/videodev.h>
#include <linux/firmware.h>

#include "pvrusb2.h"
#include "pvrusb2-v4l2.h"
#include "pvrusb2-remote.h"
#include "pvrusb2-saa7115.h"
#include "pvrusb2-encoder.h"
#include "pvrusb2-msp34x5g.h"
#include "pvrusb2-tuner.h"
#include "pvrusb2-eeprom.h"
#include "pvrusb2-buffer.h"
#include "pvrusb2-stream.h"

#define DRIVER_VERSION "0.6(08/09/2004)"
#define DRIVER_AUTHOR "Aurelien Alleaume <slts@free.fr>"
#define DRIVER_DESC "Hauppauge WinTV-PVR-USB2 MPEG2 Encoder/Tuner"

static int video_nr = -1;       // Number of Video Devices Initialized
static int pvr_nr = 0;          // sequential number of pvrusb2 device
static int freq = 0;
static int hfull = 1;
static int width = 720;

module_param(video_nr, int, -1);
MODULE_PARM_DESC(video_nr, "Offset for device's minor");
module_param(freq, int, 0);
MODULE_PARM_DESC(freq, "initial frequencie");
module_param(hfull, int, 1);
MODULE_PARM_DESC(hfull, "full height video ?");
module_param(width, int, 720);
MODULE_PARM_DESC(width, "video width : 720,480,352");

static struct usb_device_id pvr_table[] = {
	{ USB_DEVICE(0x2040, 0x2900) },
	{ }
};

static void pvr_reset_ctl_endpoints(struct pvr *pvr)
{
    usb_settoggle(pvr->udev, PVR_CTL_WRITE_ENDPOINT & 0xf,
                  !(PVR_CTL_WRITE_ENDPOINT & USB_DIR_IN), 0);
    usb_settoggle(pvr->udev, PVR_CTL_READ_ENDPOINT & 0xf,
                  !(PVR_CTL_READ_ENDPOINT & USB_DIR_IN), 0);
    usb_clear_halt(pvr->udev, usb_rcvbulkpipe(pvr->udev,
                                             PVR_CTL_READ_ENDPOINT & 0x7f));
    usb_clear_halt(pvr->udev, usb_sndbulkpipe(pvr->udev,
                                             PVR_CTL_WRITE_ENDPOINT & 0x7f));
}

int pvr_send_request(struct pvr *pvr, void *data,
                            unsigned int dlen, void *res,
                            unsigned int rlen)
{
    int ret = 0, actual_dlen;
    unsigned int pipe;


    if (pvr->disconnecting || pvr->dev_state == _unstable)
        return -EIO;

    /* send data if any. */
    if(dlen && data){
        pipe = usb_sndbulkpipe(pvr->udev, PVR_CTL_WRITE_ENDPOINT);
        ret = usb_bulk_msg(pvr->udev, pipe, data, dlen, &actual_dlen, HZ / 4);
        if (ret){
            info("transaction failed, retrying"); 
            pvr_reset_ctl_endpoints(pvr);
            /* retry */
            ret = usb_bulk_msg(pvr->udev, pipe, data, dlen, &actual_dlen, HZ / 4);
            if (ret)
                pvr_state_change(pvr, _unstable);
        }
    }

    if(rlen == 0)
        return 0;

    pipe = usb_rcvbulkpipe(pvr->udev, PVR_CTL_READ_ENDPOINT);
    ret = usb_bulk_msg(pvr->udev, pipe, res, rlen, &actual_dlen, HZ / 4);

    if (ret){
        info("transaction failed, retrying"); 
        pvr_reset_ctl_endpoints(pvr);
        /* retry */
        ret = usb_bulk_msg(pvr->udev, pipe, res, rlen, &actual_dlen, HZ / 4);
        if (ret)  
            pvr_state_change(pvr, _unstable);
    }

    return ret;
}

/* 
 * pvr_send_sequence().
 *
 * executes a sequence of commands ended by a zero length one.
 *
 */

int pvr_send_sequence(struct pvr *pvr, struct pvr_cmd *cmd)
{
    int ret = 0, i;

    i = 0;
    while ((cmd[i].length != 0) && (ret == 0)){
        memcpy(pvr->cmd_buffer, cmd[i].cmd, cmd[i].length);
        ret |= pvr_send_request(pvr, pvr->cmd_buffer, cmd[i].length,
                                pvr->cmd_buffer, cmd[i].ret);
        i++;
    }
    return ret;
}

/* 
 * pvr_upload_firmware1().
 *
 * It will upload the firmware to the inboard USB controller (FX2) ; Making
 * it re-enumerate.
 *
 * NOTE : the pointer to the firmware data given by request_firmware()
 * is not suitable for an usb transaction.
 *
 */

static int pvr_upload_firmware1(struct usb_device *udev, char *firmware)
{
    const struct firmware *fw_entry = 0;
    void  *fw_ptr;
    unsigned int pipe;
    int ret;
    u16 address;

    pvr_debug("pvr_upload_firmware1");

    usb_settoggle(udev, 0 & 0xf, !(0 & USB_DIR_IN), 0);
    usb_clear_halt(udev, usb_sndbulkpipe(udev, 0 & 0x7f));

    pipe = usb_sndctrlpipe(udev, 0);
    ret = request_firmware(&fw_entry, firmware, &udev->dev);
    
    if (ret) {
        err("request_firmware failed for '%s'", firmware);
        return ret;
    }

    if (fw_entry->size != 0x2000){
        err("wrong fx2 firmware size");
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    fw_ptr = kmalloc(0x800, GFP_KERNEL);
    if (fw_ptr == NULL){
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    /*
     * Write 0x01 to the CPUCS register at address e600 to stop the CPU 
     * during firmware upload.
     */
    
    ((u8*) fw_ptr)[0] = 0x01;
    ret = usb_control_msg(udev, pipe, 0xa0, 0x40, 0xe600, 0, fw_ptr, 1, HZ);

    /*
     * upload the firmware to address 0000-1fff in 2048 (=0x800) bytes chunk.
     */

    for(address = 0; address < fw_entry->size; address += 0x800){ 
        memcpy(fw_ptr, fw_entry->data + address, 0x800);
        ret += usb_control_msg(udev, pipe, 0xa0, 0x40, address,
                               0, fw_ptr, 0x800, HZ);
    }

    /*
     * Release the CPU from reset state by writing 0x00 to CPUCS.
     */
    ((char*) fw_ptr)[0] = 0x00;
    ret += usb_control_msg(udev, pipe, 0xa0, 0x40, 0xe600, 0, fw_ptr, 1, HZ);
    pvr_debug("CPUCS register restart command : %d byte written", ret);

    kfree(fw_ptr);
    release_firmware(fw_entry);

    /* We should have written 1 + 8192 + 1 bytes */
    return (ret == 8194) ? 0 : -EIO;
}

int pvr_write_register(struct pvr *pvr, u16 reg, u32 data)
{
    int ret;

//    pvr_debug("pvr_write_register");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = 0x04;  /* write register prefix */
    PVR_DECOMPOSE_BE(pvr->cmd_buffer,1,data);
    pvr->cmd_buffer[5] = 0;
    pvr->cmd_buffer[6] = (reg >> 8) & 0xff;
    pvr->cmd_buffer[7] = reg & 0xff;


    ret = pvr_send_request(pvr, pvr->cmd_buffer, 8, pvr->cmd_buffer, 0);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_read_register(struct pvr *pvr, u16 reg, u32 *data)
{
    int ret = 0;

//    pvr_debug("pvr_read_register");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = 0x05;  /* read register prefix */
    pvr->cmd_buffer[1] = 0;
    pvr->cmd_buffer[2] = 0;
    pvr->cmd_buffer[3] = 0;
    pvr->cmd_buffer[4] = 0;
    pvr->cmd_buffer[5] = 0;
    pvr->cmd_buffer[6] = (reg >> 8) & 0xff;
    pvr->cmd_buffer[7] = reg & 0xff;

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, 8, pvr->cmd_buffer, 4);
    *data = *((u32 *)pvr->cmd_buffer) & 0xffff;

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_write_u16(struct pvr *pvr, u16 data, int res)
{
    int ret;

//    pvr_debug("pvr_write_u16");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = (data >> 8) & 0xff;
    pvr->cmd_buffer[1] = data & 0xff;

    ret = pvr_send_request(pvr, pvr->cmd_buffer, 2, pvr->cmd_buffer, res);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_write_u8(struct pvr *pvr, u8 data, int res)
{
    int ret;

//    pvr_debug("pvr_write_u8");

    down(&pvr->ctl_sem);

    pvr->cmd_buffer[0] = data;

    ret = pvr_send_request(pvr, pvr->cmd_buffer, 1, pvr->cmd_buffer, res);

    up(&pvr->ctl_sem);

    return ret;
}

int pvr_i2c_write(struct pvr *pvr, u8 i2c_addr, u8 *data, u8 length)
{
    int ret = 0;

//    pvr_debug("pvr_i2c_write");

    down(&pvr->ctl_sem);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);

    pvr->cmd_buffer[0] = 0x08;  /* write prefix */
    pvr->cmd_buffer[1] = i2c_addr;  /* i2c addr of chip */
    pvr->cmd_buffer[2] = length;  /* length of what follows */

    memcpy(pvr->cmd_buffer + 3, data, length);

    ret |= pvr_send_request(pvr, pvr->cmd_buffer, length + 3, pvr->cmd_buffer, 1);
    /* FIXME : we might want to check pvr->cmd_buffer[0] for write success  (08 ?) */
    up(&pvr->ctl_sem);

    return ret;
}

int pvr_i2c_read(struct pvr *pvr, u8 i2c_addr, u8 *data, u8 dlen, u8 *res, u8 rlen)
{
    int ret;

//    pvr_debug("pvr_i2c_read");

    down(&pvr->ctl_sem);

    memset(pvr->cmd_buffer, 0, PVR_CTL_BUFFSIZE);
    
    pvr->cmd_buffer[0] = 0x09;  /* read prefix */
    pvr->cmd_buffer[1] = dlen;  /* arg length */
    pvr->cmd_buffer[2] = rlen;  /* anwer length. Device will send one more byte (status). */
    pvr->cmd_buffer[3] = i2c_addr;  /* i2c addr of chip */

    if (data && dlen)
        memcpy(pvr->cmd_buffer + 4, data, dlen);

    ret = pvr_send_request(pvr, pvr->cmd_buffer, 4 + dlen, pvr->cmd_buffer, rlen + 1);

    if (res)
        memcpy(res, pvr->cmd_buffer + 1, rlen);

    up(&pvr->ctl_sem);

    return ret;
}

/*
 * pvr_do_command()
 *
 * Executes a simple command.
 */

int pvr_do_command(struct pvr *pvr, struct pvr_cmd *cmd)
{
    int ret;

    pvr_debug("pvr_do_command");

    if (cmd == NULL) 
        return -EIO;

    down(&pvr->ctl_sem);

    /* data command (cmd->cmd) is (often) statically allocated. For usb
     *  transfer we need dynamic memory.
    */
    memcpy(pvr->cmd_buffer, cmd->cmd, cmd->length);

    ret = pvr_send_request(pvr, pvr->cmd_buffer, cmd->length,
                           pvr->cmd_buffer, cmd->ret);

    up(&pvr->ctl_sem);

    return ret;
}

static int pvr_configure_capture(struct pvr *pvr)
{
    int ret = 0;

    ret |= pvr_configure_decoder(pvr);

    ret |= pvr_configure_encoder(pvr);

    ret |= pvr_configure_tuner(pvr);

    ret |= pvr_configure_audio(pvr);
    
    ret |= pvr_decoder_set_input(pvr);
    
    return ret;
}

/*
 * pvr_stop()
 *
 * Stop the stream. Call with lock held. After it, we can free all the buffer
 * without any danger, as long as we don't release the pvr->sem lock.
 */

static void pvr_stop(struct pvr *pvr)
{
    pvr_debug("pvr_stop");

    /* mask all interrupts */
    pvr_write_register(pvr, 0x0048, 0xffffffff);
    /* stop capture on encoder side */
    pvr_encoder_stop(pvr);
    /* stop capture */
    pvr_write_u8(pvr, 0x37, 0);
    pvr_write_u8(pvr, 0x44, 0);
    /* change some GPIO data */
    pvr_write_register(pvr, 0x9020, 0x81040000);
    pvr_write_register(pvr, 0x900c, 0x08000000);
}

/*
 * pvr_start()
 *
 * Start the stream. Call with lock held.
 *
 */

PVR_DECLARE_CMD(start_capture, 0, "\x36\x03\x00\x00\x00\x00\x00\x44");

static void pvr_start(struct pvr *pvr)
{
    pvr_debug("pvr_start");

    /* unmask some interrupts */
    pvr_write_register(pvr, 0x0048, 0xbfffffff);
    /* change some GPIO data */
    pvr_write_register(pvr, 0x9020, 0x80040000);
    pvr_write_register(pvr, 0x900c, 0x00000000);

    /* Start captutre (in MPEG) */
//    pvr_do_command(pvr, &start_capture);
    pvr_write_u8(pvr, 0x36, 0);
    pvr_write_u8(pvr, 0x43, 0);
    pvr_encoder_start(pvr);
}

/* State changing functions. Call with lock held. */

void pvr_state_change(struct pvr *pvr, enum pvr_device_state new_state)
{
    enum pvr_device_state old_state = pvr->dev_state;

    if ((old_state == _reset) && (new_state == _reset)){
        pvr->dev_state = _unstable;
        return;
    }

    if (old_state == new_state || old_state == _unstable)
        return;

    /* going to a lower state */

    if (new_state == _unstable)
        /* prevent any ctl io */
        pvr->dev_state = _unstable;

    if (new_state == _reset) {
        pvr_reset_ctl_endpoints(pvr);
        /* prevent recursive reset */
        pvr->dev_state = _reset;
    }

    if ((new_state <= _stopped) && (old_state > _stopped))
        pvr_stop(pvr);

    if (new_state <= old_state)
        goto done;

    /* changing to an higher state */

    switch (old_state){
        case _reset:
        case _need_firmware:
            if (new_state == _need_firmware)
                break;
            /* As it sometimes fails, give it 2 tries */
            if (pvr_upload_firmware2(pvr) && pvr_upload_firmware2(pvr)){
                err("device unstable!!");
                pvr->dev_state = _unstable;
                return;
            }
        case _need_config:
            if (new_state == _need_config)
                break;
            pvr_configure_capture(pvr);
        case _stopped:
            if (new_state == _stopped)
                break;
            pvr_start(pvr);
        default:
            break;
    }

done:
    pvr->dev_state = new_state;

}

void pvr_state_above(struct pvr *pvr, enum pvr_device_state state)
{
    if (pvr->dev_state < state)
        pvr_state_change(pvr, state);
}

void pvr_state_below(struct pvr *pvr, enum pvr_device_state state)
{
    if (pvr->dev_state > state) 
        pvr_state_change(pvr, state);
}

void pvr_reconfigure(struct pvr *pvr)
{
    enum pvr_device_state old_state = pvr->dev_state;

    pvr_state_below(pvr, _need_config);
    pvr_state_above(pvr, old_state);

}

int pvr_config_apply(struct pvr *pvr, enum pvr_config config)
{
    pvr_debug("pvr_config_apply");

    pvr_state_below(pvr, _need_config);
    pvr->config = config;
    pvr_state_above(pvr, _streaming);

    return 0;
}

static void pvr_configure_pvr(struct pvr *pvr)
{

#define PVR_DEFAULT_REG(id) (pvr_regular_ctl[id-V4L2_CID_BASE].default_value)
#define PVR_DEFAULT_PRIV(id) (pvr_private_ctl[id-V4L2_CID_PRIVATE_BASE].default_value)

    pvr->video.standard = PVR_STD_SECAM;
    pvr->video.input = 0;
    pvr->audio.input = 0;
    pvr->video.width = width;
    pvr->video.height_full = hfull;
    pvr->tv_frequency = freq;

    pvr->video.bitrate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_VIDEOBITRATE);
    pvr->video.vbr = PVR_DEFAULT_PRIV(V4L2_CID_PVR_VBR);
    pvr->video.peakbitrate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_VIDEOPEAK);
    pvr->video.brightness = PVR_DEFAULT_REG(V4L2_CID_BRIGHTNESS);
    pvr->video.contrast = PVR_DEFAULT_REG(V4L2_CID_CONTRAST);
    pvr->video.saturation = PVR_DEFAULT_REG(V4L2_CID_SATURATION);
    pvr->video.hue = PVR_DEFAULT_REG(V4L2_CID_HUE);
    pvr->audio.mode = V4L2_TUNER_MODE_STEREO;
    pvr->audio.bitrate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_AUDIOBITRATE);
    pvr->audio.srate = PVR_DEFAULT_PRIV(V4L2_CID_PVR_SRATE);
    pvr->audio.crc = PVR_DEFAULT_PRIV(V4L2_CID_PVR_AUDIOCRC);
    pvr->audio.emphasis = PVR_DEFAULT_PRIV(V4L2_CID_PVR_AUDIOEMPHASIS);
    pvr->audio.layer = 2;
    pvr->audio.lang = 1;
    pvr->audio.muted = PVR_DEFAULT_REG(V4L2_CID_AUDIO_MUTE);
    pvr->audio.volume = PVR_DEFAULT_REG(V4L2_CID_AUDIO_VOLUME);
    pvr->audio.balance = PVR_DEFAULT_REG(V4L2_CID_AUDIO_BALANCE);
    pvr->audio.bass = PVR_DEFAULT_REG(V4L2_CID_AUDIO_BASS);
    pvr->audio.treble = PVR_DEFAULT_REG(V4L2_CID_AUDIO_TREBLE);
}

static int pvr_probe(struct usb_interface *intf, const struct usb_device_id *devid)
{
	struct usb_device *udev = interface_to_usbdev(intf);
	__u8 ifnum = intf->cur_altsetting->desc.bInterfaceNumber;
	struct pvr *pvr = NULL;
    int i, ret = 0;

    info("endpoints : %i",intf->cur_altsetting->desc.bNumEndpoints);

    usb_set_interface(udev, ifnum, 0);

 	pvr = kmalloc(sizeof(struct pvr), GFP_KERNEL);
    if (pvr == NULL) {
		err("%s: couldn't allocate pvr struct", __FUNCTION__);
		return -ENOMEM;
	}

    memset(pvr, 0, sizeof(struct pvr));

    /*
     * If the FX2 firmware is not loaded, do it. Then the device will
     * re-enumerate, closing this driver instance.
     */
    if (intf->cur_altsetting->desc.bNumEndpoints == 0) {
        pvr->renum = 1;
        usb_set_intfdata(intf, pvr);
        pvr_upload_firmware1(udev, FIRMWARE1_FILE);
        return 0;
    }

    pvr->udev = udev;
    init_MUTEX(&pvr->sem);
    init_MUTEX(&pvr->ctl_sem);

    pvr->nr = pvr_nr++;
    pvr->dev_state = _need_firmware;

    pvr_configure_pvr(pvr);

    ret = pvr_input_init(pvr);
    if (ret)
        goto err0;

    usb_set_intfdata(intf, pvr);
    pvr_reset_ctl_endpoints(pvr);
    pvr_eeprom_analyse(pvr);
    pvr_input_start(pvr);

    /* regiter streams */

    down(&pvr->sem);

    pvr->stream[0] = pvr_stream_alloc(pvr, PVR_VID_ENDPOINT,
                                      VFL_TYPE_GRABBER,video_nr, _mpeg);
    pvr->stream[1] = pvr_stream_alloc(pvr, PVR_VBI_ENDPOINT,
                                      VFL_TYPE_VBI, video_nr, _vbi);
    pvr->stream[2] = pvr_stream_alloc(pvr, PVR_VID_ENDPOINT,
                                      VFL_TYPE_RADIO, video_nr, _radio);

    pvr->stream_count = 3;

    for (i = 0; i < pvr->stream_count ; i++)
        if (pvr->stream[i] == NULL){
            ret = -ENOMEM;
            break;
        }

    if (ret) {
        /* One stream failed to alloc. Free'em all */
        for (i = 0 ; i < pvr->stream_count ; i++)
            if (pvr->stream[i])
                pvr_stream_free(pvr->stream[i]);
        goto err1;
    }
    up(&pvr->sem);

    return 0;

err1:
    up(&pvr->sem);
    pvr_input_stop(pvr);
    pvr_input_clear(pvr);
err0:
    kfree(pvr);

    return -ENOMEM;
        
}

/*
 * pvr_disconnect()
 *
 */

static void pvr_disconnect(struct usb_interface *intf)
{
    struct pvr *pvr = usb_get_intfdata(intf);
    int i;

    pvr_debug("pvr_disconnect");
// FIXME
    pvr->disconnecting = 1;

    if(pvr->renum){
        /* we did not initialize everything because device was 
         * going to re-enumerate
         */
        kfree(pvr);
        return;
    }

    down(&pvr->sem);

    /* Stop remote stuff. */
    pvr_input_stop(pvr);
    pvr_input_clear(pvr);

    usb_set_intfdata (intf, NULL);

    for (i = 0 ; i < pvr->stream_count ; i++)
        pvr_stream_free(pvr->stream[i]);

    up(&pvr->sem);
    kfree(pvr);
}

static struct usb_driver pvr_driver = {
    owner:   THIS_MODULE,
    name:           "pvrusb2",
    id_table:       pvr_table,
    probe:          pvr_probe,
	disconnect:     pvr_disconnect
};

/*
 * pvr_init() / pvr_exit()
 *
 * This code is run to initialize/exit the driver.
 *
 */

static int __init pvr_init(void)
{
	int ret;

    pvr_debug("pvr_init");
 
    ret = usb_register(&pvr_driver);

	if (ret == 0)
		info(DRIVER_DESC " : " DRIVER_VERSION);

	return ret;
}

static void __exit pvr_exit(void)
{
    
    pvr_debug("pvr_exit");

    usb_deregister(&pvr_driver);
}

module_init(pvr_init);
module_exit(pvr_exit);

MODULE_DEVICE_TABLE (usb, pvr_table);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
