/*
 *  
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2.h"
#include "pvrusb2-tuner.h"
#include "pvrusb2-v4l2.h"

#define PVR_TUNER_I2C_ADDR  0x61
#define PVR_STD_I2C_ADDR    0x43

struct pvr_tuner 
{
	char *name;
    unsigned char idx;  /* Hauppauge identifier */
	unsigned short thresh1; /* limit between VHF LOW and VHF HIGH */
	unsigned short thresh2; /* limit between VHF HIGH and UHF */
	unsigned char vhf_l;
	unsigned char vhf_h;
	unsigned char uhf;
	unsigned char config_tv;
	unsigned char config_radio;
	unsigned short IFPCoff; /* 622.4=16*38.90 MHz PAL, 
				   732  =16*45.75 NTSCi, 
				   940  =58.75 NTSC-Japan */
};

/*
 * Tuner found in pvrusb2 devices. 
 * NOTE : This data is not exactly the same as kernel tuner.c one. Why ?
 */

#define PVR_NTUNER  4

struct pvr_tuner pvr_tuners[] = {
    { "Philips PAL_I (FI1246 and compatibles)", 24,
      16*140.25, 16*463.25, 0xa0, 0x90, 0x30, 0xce, 0xc8, 623},
    { "LG PAL_BG+FM (TPI8PSB01D)", 47,
      16*170, 16*450, 0xa0, 0x90, 0x30, 0xce, 0xc8, 623 },
    { "Microtune 4049 FM5", 52,
      16*141, 16*464, 0xa0, 0x90, 0x30, 0xce, 0xc8, 623 },
    { "Philips NTSC MK3 (FM1236MK3 or FM1236/F)", 58,
      16*160, 16*442, 0x01, 0x02, 0x04, 0xce, 0xc8, 732 },
};

void pvr_set_tuner(struct pvr *pvr, int idx)
{
    int i;

    pvr->tuner = 0;

    for (i=0; (i < PVR_NTUNER) && (idx != pvr_tuners[i].idx); i++);

    if (idx == pvr_tuners[i].idx)
        pvr->tuner = i;
    else
        err("unknow tuner %i. Please contact developper.", idx);
}

int pvr_set_standard(struct pvr *pvr)
{
    u8 buff [4]; 

    pvr_debug("pvr_set_standard");

    PVR_BUG_ON(pvr->video.standard < 0 || pvr->video.standard >= PVR_V4L2_NSTD);

    /* FIXME: Might be tuner dependant ? Need more hardware info... */
    PVR_DECOMPOSE_BE(buff, 0, pvr_tbl_standard[pvr->video.standard]);

    return pvr_i2c_write(pvr, PVR_STD_I2C_ADDR, buff, 4);
}

int pvr_set_freq_tv(struct pvr *pvr)
{
    u32 val;
    u8 buff[4];

    pvr_debug("pvr_set_freq_tv : %i", pvr->tv_frequency);

    PVR_BUG_ON(pvr->tuner < 0 || pvr->tuner >= PVR_NTUNER);

    val = pvr_tuners[pvr->tuner].IFPCoff + pvr->tv_frequency;

    buff[0] = (val >> 8) & 0xff;
    buff[1] = val & 0xff;
    buff[2] = pvr_tuners[pvr->tuner].config_tv;

    if (pvr->tv_frequency < pvr_tuners[pvr->tuner].thresh1)
        /* VHF LOW */
        buff[3] = pvr_tuners[pvr->tuner].vhf_l;
    else if (pvr->tv_frequency < pvr_tuners[pvr->tuner].thresh2) 
        /* VHF HIGH */
        buff[3] = pvr_tuners[pvr->tuner].vhf_h;
    else
        /* UHF */
        buff[3] = pvr_tuners[pvr->tuner].uhf;

    return pvr_i2c_write(pvr, PVR_TUNER_I2C_ADDR, buff, 4);
}

int pvr_set_freq_radio(struct pvr *pvr)
{
    u16 val;
    u8 buff[4];

    pvr_debug("pvr_set_freq_radio : %i", pvr->radio_frequency);

    PVR_BUG_ON(pvr->tuner < 0 || pvr->tuner >= PVR_NTUNER);

    val = (int)(16*10.7) + pvr->radio_frequency;

    buff[0] = (val >> 8) & 0xff;
    buff[1] = val & 0xff;
    buff[2] = pvr_tuners[pvr->tuner].config_radio;
    buff[3] = 0xa6;

    return pvr_i2c_write(pvr, PVR_TUNER_I2C_ADDR, buff, 4);
}

int pvr_configure_tuner(struct pvr *pvr)
{
    int ret = 0;

    ret |= pvr_set_standard(pvr);
    if (pvr->config == _radio)
        ret |= pvr_set_freq_radio(pvr);
    else
        ret |= pvr_set_freq_tv(pvr);

    return ret;
}
